import seqio
from typing import Optional
from pydantic import BaseModel

from seqio.vocabularies import Vocabulary
from tensorflow.python.framework.dtypes import (
    int32,
    DType
)
from tensorflow.python.platform.gfile import GFile
from sentencepiece import sentencepiece_model_pb2 as sentencepiece_model
from tfdsio.utils import logger

class ModelVocab(BaseModel):
    """
    Base Model for SPE Model
    """
    path: str
    size: int
    extra_ids: int = 100


_T5Vocabs = {
    'default': ModelVocab(
        path = 'gs://t5-data/vocabs/cc_all.32000.100extra/sentencepiece.model', 
        size = 32000, 
        extra_ids = 100
    ),
    'mt5': ModelVocab(
        path = 'gs://t5-data/vocabs/mc4.250000.100extra/sentencepiece.model',
        size = 250000,
        extra_ids = 100
    )
}

class Vocabs:
    
    @classmethod
    def get_vocabulary(
        cls, 
        name: Optional[str] = 'mt5',
        path: Optional[str] = None,
        size: Optional[int] = None,
        extra_ids: Optional[int] = None,
    ) -> Vocabulary:
        """
        Get a T5 Vocabulary
        """
        return _T5Vocabs.get(name, ModelVocab(path = path, size = size, extra_ids = extra_ids))

    @classmethod
    def load_sentencepiece_model(cls, vocab_path: str):
        """
        Load a SentencePiece Model
        """
        sp_model = sentencepiece_model.ModelProto()
        sp_model.ParseFromString(GFile(vocab_path).read())
        return sp_model

    @classmethod
    def calculate_sentencepiece_extra_ids(cls, vocab_path: str, show_extra: bool = False):
        """
        Calculate the number of extra IDs for a SentencePiece Model
        """
        sp_model = cls.load_sentencepiece_model(vocab_path)
        vocab_size = len(sp_model.pieces)
        extra_ids = 0 if vocab_size % 1000 == 0 else 1000 - (vocab_size % 1000)
        base_size = vocab_size - extra_ids
        if extra_ids and show_extra:
            logger.info(f'Extra IDs: {extra_ids}')
            for i, piece in enumerate(sp_model.pieces[base_size:]):
                logger.info(f'[{i+base_size}] -> {piece.piece}')

        logger.info(f'Extra IDs: {extra_ids}. SPE Model Size: {vocab_size}')
        return extra_ids


    @classmethod
    def get_output_features(
        cls,
        input_key: str = 'inputs', 
        target_key: str= 'targets',
        vocab: Optional[ModelVocab] = None,
        vocab_path: Optional[str] = None,
        vocab_name: Optional[str] = None,
        inputs_add_eos: bool = True,
        targets_add_eos: bool = True,
        inputs_required: bool = False, # For LM, Inputs are not required
        targets_required: bool = True,
        input_dtype: Optional[DType] = int32,
        target_dtype: Optional[DType] = int32,

        extra_ids: Optional[int] = None,
        reverse_extra_ids: Optional[bool] = False,

    ):
        if vocab: 
            vocab_path = vocab.path
            extra_ids = vocab.extra_ids
        
        elif vocab_name:
            vocab = cls.get_vocabulary(vocab_name)
            vocab_path = vocab.path
            extra_ids = vocab.extra_ids
        
        elif vocab_path and extra_ids is None:
            extra_ids = cls.calculate_sentencepiece_extra_ids(vocab_path)

        return {
            input_key: seqio.Feature(
                seqio.SentencePieceVocabulary(
                    vocab_path, 
                    extra_ids = extra_ids,
                    reverse_extra_ids = reverse_extra_ids,
                ),
                add_eos = inputs_add_eos, 
                required = inputs_required, 
                dtype = input_dtype
            ),
            target_key: seqio.Feature(
                seqio.SentencePieceVocabulary(
                    vocab_path, 
                    extra_ids = extra_ids,
                    reverse_extra_ids = reverse_extra_ids,
                ),
                add_eos = targets_add_eos, 
                required = targets_required,
                dtype = target_dtype
            ),
        }



import os
import csv
import json
import termcolor
import dataclasses
import posixpath
import numpy as np
from seqio.vocabularies import Vocabulary
from tensorflow.python.framework.dtypes import (
    int32,
    int64,
    DType
)
#import multiprocessing as mp
import tensorflow_datasets as tfds
from tensorflow_datasets.core import constants
from tensorflow_datasets.core import file_adapters
from tensorflow_datasets.core.utils import file_utils
from tensorflow_datasets.core.utils import Version, version
from tensorflow_datasets.core.download import DownloadManager

from tensorflow.python.platform import gfile
from tensorflow.python.platform.gfile import GFile
from tensorflow.python.data.ops.dataset_ops import AUTOTUNE
from tensorflow.python.data.ops.readers import TextLineDataset
from tensorflow.python.data.ops.readers import TFRecordDataset
from tensorflow.python.framework.ops import enable_eager_execution, disable_eager_execution

from typing import Union, Dict, Any, Optional, List, ClassVar, Callable
from tfdsio.utils import logger
from tfdsio.io import (
    PipelineIO
)
try:
    from datasets import Dataset, DatasetDict
except ImportError:
    Dataset = object
    DatasetDict = object


_has_logged_ds_info = set()


def log_dataset_info(name, msg):
    global _has_logged_ds_info
    if name in _has_logged_ds_info:
        return
    termcolor.cprint(msg, attrs=["bold"])
    _has_logged_ds_info.add(name)

@dataclasses.dataclass(frozen=True)
class Feature:
    """A container for attributes of output features of data providers."""
    vocabulary: Vocabulary
    add_eos: bool = True
    required: bool = True
    dtype: DType = int32


@dataclasses.dataclass(frozen=True)
class ShardInfo:
    """A container for specifying sharding info."""
    index: int
    num_shards: int

_DataFeatures: Dict[str, Any] = {
    'text': tfds.features.Text(),
    'image': tfds.features.Image(),
    'label': tfds.features.ClassLabel(),
    'audio': tfds.features.Audio(),
    #'sequence': tfds.features.Sequence(tfds.features.Text()),
}

_FilePipelines: Dict[str, Callable] = {
    'text': PipelineIO.get_text_reader,
    'textlines': PipelineIO.get_textline_reader,
    'csv': PipelineIO.get_csv_reader,
    'json': PipelineIO.get_json_reader,
    'jsonlines': PipelineIO.get_jsonlines_reader,
}

_DataSplits: Dict[str, tfds.Split] = {
    'train': tfds.Split.TRAIN,
    'validation': tfds.Split.VALIDATION,
    'test': tfds.Split.TEST,
}

class DataFeatures:

    @classmethod
    def get_feature(cls, name: str):
        return _DataFeatures.get(name, _DataFeatures['text'])

    @classmethod
    def to_features(
        cls,
        features: Dict[str, Any]
    ) -> tfds.features.FeaturesDict:
        """
        Convert a Dictionary of Features to a FeaturesDict
        """
        return tfds.features.FeaturesDict({
            k: cls.get_feature(v) for k, v in features.items()
        })


# class DataSplits:
#     train = tfds.Split.TRAIN
#     validation = tfds.Split.VALIDATION
#     test = tfds.Split.TEST


VersionOrStr = Union[Version, str]


@dataclasses.dataclass
class BuilderConfig:
    name: Optional[str] = None
    classifier: Optional[str] = None
    version: Optional[VersionOrStr] = '1.0.0'
    release_notes: Optional[Dict[str, str]] = None
    supported_versions: List[str] = None
    description: Optional[str] = ''
    
    dataset_urls: Optional[Union[str, List[str], Dict[str, Union[str, List[str]]]]] = None
    file_format: Optional[str] = None
    dataset_format: Optional[str] = 'jsonlines'
    features: Optional[Any] = None
    datamap: Optional[Any] = None
    supervised_keys: Optional[Any] = None
    homepage: Optional[str] = ''
    citation: Optional[Any] = None
    metadata: Optional[Any] = None
    redistribution_info: Optional[Any] = None
    
    data_dir: Optional[str] = None
    stream_data: bool = True # Stream data if from GCS / FileLike without Downloading
    preprocessors: Optional[List[Callable]] = None
    preprocessor_kwargs: Optional[Dict[str, Any]] = None
    
    # Will re-process the dataset
    rewrite_existing: bool = False

    def get_datasources(
        self, 
        dataset_urls: Optional[Union[str, List[str], Dict[str, Union[str, List[str]]]]]
    ) -> Dict[str, Union[str, List[str]]]:

        if isinstance(dataset_urls, str):
            return {'train': [dataset_urls]}
        if isinstance(dataset_urls, list):
            return {'train': dataset_urls}
        
        dataset_src = {'extra': []}
        for k, v in dataset_urls.items():
            if k in {'train', 'training'}:
                dataset_src['train'] = v
            elif k in {'dev', 'develop', 'development', 'val', 'validation', 'eval', 'evaluation'}:
                dataset_src['validation'] = v
            elif k in {'test', 'gold'}:
                dataset_src['test'] = v
            elif isinstance(v, str):
                dataset_src['extra'].append(v)
            elif isinstance(v, list):
                dataset_src['extra'].extend(v)
            else:
                dataset_src['extra'] = v
        
        for k, v in dataset_src.items():
            if isinstance(v, str):
                dataset_src[k] = [v]
        return dataset_src

    def parse_config(self):
        if self.features: self.features = DataFeatures.to_features(self.features)
        if self.supported_versions is None: self.supported_versions = []
        if self.dataset_urls: self.dataset_urls = self.get_datasources(self.dataset_urls)
        if self.dataset_format: assert self.dataset_format in _FilePipelines, f"Invalid dataset format: {self.dataset_format}. Supported formats: {_FilePipelines.keys()}"

    def from_config(self, config: Dict[str, Any]):
        keys = list(self.__dict__.keys())
        for k,v in config.items():
            if k in keys: self.__dict__[k] = v
        self.parse_config()

    def from_file(self, filename):
        assert gfile.Exists(filename)
        with GFile(filename, 'r') as f:
            config = json.load(f)
        self.from_config(config)
    
    def from_auto(self, config_or_file: Union[Dict[str, Any], str]):
        if isinstance(config_or_file, dict):
            return self.from_config(config_or_file)
        return self.from_file(config_or_file)
    
    @property
    def dataset_info(self):
        keys = ['description', 'features', 'supervised_keys', 'homepage', 'citation', 'metadata', 'redistribution_info']
        return {k: self.__dict__[k] for k in keys}
    
    @property
    def dataset_name(self):
        return f'{self.name}/{self.classifier}:{self.version}'


class DatasetInfo(tfds.core.DatasetInfo):
    @property
    def full_name(self):
        """Full canonical name: (<dataset_name>/<dataset_classifier>/<version>)."""
        names = []
        if self._builder.builder_config:
            names.extend(
                (
                    self._builder.builder_config.name, 
                    self._builder.builder_config.classifier, 
                    str(self.version)
                )
            )
        return posixpath.join(*names)


class TFDSDatasetBuilder(tfds.core.GeneratorBasedBuilder):

    VERSION = None
    RELEASE_NOTES: ClassVar[Dict[str, str]] = {}
    SUPPORTED_VERSIONS = []
    BUILDER_CONFIGS = []
    MANUAL_DOWNLOAD_INSTRUCTIONS = None

    def __init__(
        self, 
        *, 
        config: Union[None, str, BuilderConfig] = None, 
        **kwargs
    ):
        if config.data_dir:
            config.data_dir = os.fspath(config.data_dir)  # Pathlib -> str
        elif os.environ.get('TFDS_DIR', None):
            config.data_dir = os.fspath(os.environ['TFDS_DIR'])
        elif os.environ.get('GCS_DIR', None):
            config.data_dir = os.fspath(os.environ['GCS_DIR'])
        
        file_format = config.file_format or file_adapters.DEFAULT_FILE_FORMAT
        self._original_state = dict(data_dir = config.data_dir, config = config, version = config.version)
        self._builder_config: BuilderConfig = self._create_builder_config(config)
        self.name = self._builder_config.name
        self._version = self._pick_version(config.version)
        self._data_dir_root, self._data_dir = self._build_data_dir(config)
        if config.rewrite_existing and gfile.Exists(self._data_dir):
            self._remove_existing_data_dir()
        
        if gfile.Exists(self._data_dir):
            self.info.read_from_directory(self._data_dir)
            self._log_dataset_info()
        else:
            self.info.initialize_from_bucket()
        
        try:
            self._file_format = file_adapters.FileFormat(file_format)
            self.info.set_file_format(self._file_format)

        except ValueError as e:
            all_values = [f.value for f in file_adapters.FileFormat]
            raise ValueError(f"{file_format} is not a valid format. Valid file formats: {all_values}") from e

    def _remove_existing_data_dir(self):
        """
        Removes a previously created dataset
        if config.rewrite_existing is True
        """
        logger.warning(f"Removing and overwriting dataset {self.name} at {self._data_dir}")
        gfile.DeleteRecursively(self._data_dir)
    
    def _log_dataset_info(self):
        msg = "--------------------------------------------\n"
        msg += f"TFDSIO Dataset: {self.info.full_name}\n"
        msg += "Splits:\n"
        for split, details in self.info.splits.items():
            msg += f" - {split}: {details.num_examples} examples, {details.num_shards} shards\n"
        msg += f"Features: {str(self.info.features)}\n"
        msg += f"Download Size: {self.info.download_size}, Dataset Size: {self.info.dataset_size}\n"
        msg += "--------------------------------------------"
        log_dataset_info(self.info.full_name, msg)
    
    def _log_download_done(self):
        msg = f"Dataset {self.info.full_name} downloaded and prepared to {self._data_dir}.\n"
        termcolor.cprint(msg, attrs=["bold"])
        self._log_dataset_info()
    
    def _log_download_bytes(self):
        termcolor.cprint(
            f"Downloading and Preparing dataset {self.info.full_name} - {self.info.download_size} "
            f"(download: {self.info.download_size}, "
            f"generated: {self.info.dataset_size}, "
            f"total: {self.info.download_size + self.info.dataset_size}) "
            f"\nto {self._data_dir}...",
            attrs=["bold"],
        )
    
    def _split_generators(
        self, 
        dl_manager: DownloadManager
    ):
        dataset_urls = self.builder_config.dataset_urls.copy()
        _extra = dataset_urls.pop('extra', None)
        # sort through the dataset_urls for web urls and local/gcs files 
        # if streaming is enabled
        excluded_paths = {}
        if self.builder_config.stream_data:
            excluded_paths = {
                k: [i for i in v if not i.startswith('http')] for k,v in dataset_urls.items()
            }
            logger.info(f'Excluded paths: {excluded_paths}')
            dataset_urls = {
                k: [i for i in v if i.startswith('http')] for k,v in dataset_urls.items()
            }
            logger.info(f'Included paths to download: {dataset_urls}')
            if _extra:
                excluded_paths['extra'] = [i for i in _extra if not i.startswith('http')]
                _extra = [i for i in _extra if i.startswith('http')]
            
        
        dl_paths = dl_manager.download(dataset_urls)
        msg = f"Downloading Temp Files to:\n"
        for fs, fn in dl_paths.items():
            msg += f" - {fs}: {str(fn)}\n"
        termcolor.cprint(msg)
        if _extra:
            _extra = dl_manager.download(_extra)
        
        dataset_splits = list(dataset_urls.keys())
        # if streamign is enabled, re-merge the excluded paths
        if self.builder_config.stream_data:
            if _extra:
                _extra.extend(excluded_paths.pop('extra', []))
            for k,v in excluded_paths.items():
                dl_paths[k].extend(v)
        
        return [
            tfds.core.SplitGenerator(
                name = _DataSplits[split], 
                gen_kwargs = {
                    "paths": dl_paths[split], 
                    "extra": _extra
                }
            ) for split in dataset_splits
        ]
    
    def map_to_features(self, ex, idx: int, has_preprocessor: bool = False):
        """
        Map the resulting example to the features specified in the dataset_info.json
        """
        if isinstance(ex, dict):
            if self.builder_config.datamap:
                ex = {v: ex.get(k, '') for k,v in self.builder_config.datamap.items()}
            yield idx, ex
        
        elif isinstance(ex, (list, np.array)) and has_preprocessor:
            # The preprocessor could return a list of permutated examples
            for i in ex:
                if self.builder_config.datamap:
                    i = {v: i.get(k, '') for k,v in self.builder_config.datamap.items()}
                yield idx, i
                idx += 1
        
        elif isinstance(ex, (list, np.array)):
            ex = {self.builder_config.features[x]: ex[x] for x in enumerate(ex)}
            yield idx, ex
        
        elif isinstance(ex, (int, float, str, bytes)):
            ex = {self.builder_config.features[0]: ex}
            yield idx, ex


    def _generate_examples(
        self, 
        paths: Union[str, List[str]],
        extra = None, 
        **kwargs
    ):
        idx = 0
        pipeline = _FilePipelines[self.builder_config.dataset_format](paths)
        for ex in pipeline:
            if not self.builder_config.preprocessors:
                for (n, i) in self.map_to_features(ex, idx, has_preprocessor = False):
                    yield n, i
                    idx += 1
            else:
                for preprocessor in self.builder_config.preprocessors:
                    ex = preprocessor(idx = idx, data = ex, extra = extra, **self.get_preprocessor_kwargs(**kwargs))
                    for (n, i) in self.map_to_features(ex, idx, has_preprocessor = True):
                        yield n, i
                        idx += 1

    def _relative_data_dir(self, with_version=True):
        """Relative path of this dataset in data_dir."""
        builder_data_dir = self.name
        builder_config = self._builder_config
        if builder_config:
            if builder_config.classifier:
                builder_data_dir = os.path.join(builder_data_dir, builder_config.classifier)
            else:
                builder_data_dir = os.path.join(builder_config.name)

        return os.path.join(builder_data_dir, str(self._version)) if with_version else builder_data_dir

    def _build_data_dir(self, config):
        """Return the data directory for the current version.
        Args:
        given_data_dir: `Optional[str]`, root `data_dir` passed as
            `__init__` argument.
        Returns:
        data_dir_root: `str`, The root dir containing all datasets, downloads,...
        data_dir: `str`, The version data_dir
            (e.g. `<data_dir_root>/<ds_name>/<config>/<version>`)
        """
        given_data_dir = config.data_dir
        builder_dir = self._relative_data_dir(with_version=False)
        version_dir = self._relative_data_dir(with_version=True)
        default_data_dir = file_utils.get_default_data_dir(given_data_dir=given_data_dir)
        all_data_dirs = file_utils.list_data_dirs(given_data_dir=given_data_dir)
        all_versions = set()
        requested_version_dirs = {}
        for data_dir_root in all_data_dirs:
            full_builder_dir = os.path.join(data_dir_root, builder_dir)
            data_dir_versions = set(version.list_all_versions(full_builder_dir))
            if self.version in data_dir_versions:
                requested_version_dirs[data_dir_root] = os.path.join(data_dir_root, version_dir)
            all_versions.update(data_dir_versions)

        if len(requested_version_dirs) > 1:
            raise ValueError(f"Dataset was found in more than one directory: {requested_version_dirs.values()}. Please resolve the ambiguity by explicitly specifying `data_dir=`.")

        elif len(requested_version_dirs) == 1:  # The dataset is found once
            return next(iter(requested_version_dirs.items()))

        # No dataset found, use default directory
        data_dir = os.path.join(default_data_dir, version_dir)
        if all_versions:
            logger.warning(
                "Found a different version of the requested dataset:\n"
                "%s\n"
                "Using %s instead.",
                "\n".join(str(v) for v in sorted(all_versions)),
                data_dir
            )
        return default_data_dir, data_dir

    #def set_preprocessor(self, function):
    #    self._proc = function
    
    def set_preprocessors(self, functions: Union[Callable, List[Callable]]):
        if not self.builder_config.preprocessors:
            self.builder_config.preprocessors = []
        if isinstance(functions, list):
            self.builder_config.preprocessors.extend(functions)
        else:
            self.builder_config.preprocessors.append(functions)


    def get_preprocessor_kwargs(self, **kwargs):
        if not kwargs: kwargs = {}
        if self.builder_config.preprocessor_kwargs:
            kwargs.update(self.builder_config.preprocessor_kwargs)
        return kwargs
    
    # @property
    # def has_preprocessor(self):
    #     return bool('_proc' in self.__dict__)

    # def preprocessor(self, *args, **kwargs):
    #     # sourcery skip: dict-assign-update-to-union
    #     if self.has_preprocessor:
    #         if self.builder_config.process_args:
    #             kwargs.update(self.builder_config.process_args)
    #         return self._proc(*args, **kwargs)
    #     return None

    @property
    def libname(self):
        return 'tfdsio'

    def _info(self) -> DatasetInfo:
        return DatasetInfo(builder = self, **self.builder_config.dataset_info)



class HFDatasetBuilder(TFDSDatasetBuilder):

    VERSION = None
    RELEASE_NOTES: ClassVar[Dict[str, str]] = {}
    SUPPORTED_VERSIONS = []
    BUILDER_CONFIGS = []
    MANUAL_DOWNLOAD_INSTRUCTIONS = None

    def __init__(
        self, 
        *, 
        dataset: Union[Dataset, DatasetDict],
        config: Union[None, str, BuilderConfig] = None, 
        **kwargs
    ):
        if config.data_dir:
            config.data_dir = os.fspath(config.data_dir)  # Pathlib -> str
        elif os.environ.get('TFDS_DIR', None):
            config.data_dir = os.fspath(os.environ['TFDS_DIR'])
        elif os.environ.get('GCS_DIR', None):
            config.data_dir = os.fspath(os.environ['GCS_DIR'])
        
        self.dataset = dataset
        file_format = config.file_format or file_adapters.DEFAULT_FILE_FORMAT
        self._original_state = dict(data_dir = config.data_dir, config = config, version = config.version)
        self._builder_config: BuilderConfig = self._create_builder_config(config)
        self.__dict__['name'] = self._builder_config.name
        self._version = self._pick_version(config.version)
        self._data_dir_root, self._data_dir = self._build_data_dir(config)
        
        if config.rewrite_existing and gfile.Exists(self._data_dir):
            self._remove_existing_data_dir()
        
        if gfile.Exists(self._data_dir):
            self.info.read_from_directory(self._data_dir)
            self._log_dataset_info()
        else:
            self.info.initialize_from_bucket()
        
        try:
            self._file_format = file_adapters.FileFormat(file_format)
            self.info.set_file_format(self._file_format)

        except ValueError as e:
            all_values = [f.value for f in file_adapters.FileFormat]
            raise ValueError(f"{file_format} is not a valid format. Valid file formats: {all_values}") from e

    def map_split_name(self, split: str):
        """
        Remaps Split Names to the ones used in the dataset
        """
        #if 'train' in split:
        #    return _DataSplits['train']
        if split in {'dev', 'develop', 'development', 'val', 'validation', 'eval', 'evaluation'}:
            return _DataSplits['validation']
        if split in {'test', 'testing'}:
            return _DataSplits['test']
        # assume train
        return _DataSplits['train']


    def _split_generators(
        self, 
        dl_manager: DownloadManager
    ):
        if isinstance(self.dataset, Dataset):
            self.dataset = DatasetDict({'train': self.dataset})
        return [
            tfds.core.SplitGenerator(
                name = self.map_split_name(split), 
                gen_kwargs = {
                    'dataset': self.dataset[split],
                }
            ) for split in self.dataset
        ]


    def _generate_examples(
        self, 
        dataset: Dataset,
        **kwargs
    ):
        idx = 0
        for ex in dataset:
            for (n, i) in self.map_to_features(ex, idx, has_preprocessor = False):
                yield n, i
                idx += 1



__all__ = [
    'Feature',
    'ShardInfo',
    'Vocabulary',
    'GFile',
    'int32',
    'int64',
    'DType',
    'logger',
    'tfds',
    'DataFeatures',
    #'DataSplits',
    'VersionOrStr',
    'BuilderConfig',
    'TFDSDatasetBuilder',
]


#import tensorflow as tf
from tensorflow.python.framework.constant_op import constant
from tensorflow.python.util.compat import as_bytes
from tensorflow.python.framework.dtypes import (
    int32,
    int64,
    string,
    float32,
    float64,
    DType
)
from tensorflow.python.framework.dtypes import bool as tf_bool

from tensorflow.python.training.training import (
    BytesList,
    Feature,
    Int64List,
    Example,
    FloatList,
    Features,
)


def dict_to_tfexample(ex):
    """Convert example dictionary to tf.train.Example proto."""
    feature_dict = {}
    for k, v in ex.items():
        t = constant(v)
        if len(t.shape) == 0:  # pylint:disable=g-explicit-length-test
            v = [v]
        elif len(t.shape) == 1: v = list(v)
        else: raise ValueError("Unsupported shape (%s) for '%s' value: %s" % (t.shape, k, v))

    if t.dtype == string and len(t.shape) <= 1:
        feature_dict[k] = Feature(bytes_list = BytesList(value = [as_bytes(t) for t in v]))
    elif t.dtype in (tf_bool, int32, int64) and len(t.shape) <= 1:
        feature_dict[k] = Feature(int64_list = Int64List(value=v))
    elif t.dtype in (float32, float64) and len(t.shape) <= 1:
        feature_dict[k] = Feature(float_list = FloatList(value=v))
    else:
        raise ValueError("Unsupported type (%s) and shape (%s) for '%s' value: %s" % (t.dtype, t.shape, k, v))

    return Example(features = Features(feature=feature_dict))



# Data Sources
import abc
import inspect

from typing import Any, Callable, Iterable, Mapping, List, Optional, Sequence, Union
from tensorflow.python.data import Dataset


from .registries import DatasetProviderBase
from .base import Feature, ShardInfo
from .loaders import LazyTFDSLoader, LazyTFDSIOLoader, LazyHFTFDSLoader


class DataSource(DatasetProviderBase):
    """A `DatasetProvider` that provides raw data from an input source.
    Inherits all abstract methods and properties of `DatasetProviderBase` except
    those overidden below.
    """

    def __init__(self, splits: Iterable[str], num_input_examples: Optional[Mapping[str, int]] = None):
        self._splits = tuple(splits)
        self._num_input_examples = (dict(num_input_examples) if num_input_examples is not None else None)

    @property
    def splits(self) -> Sequence[str]:
        return self._splits

    @property
    def output_features(self) -> Mapping[str, Feature]:
        """Override unused property of `DatasetProviderBase`."""
        raise NotImplementedError

    @abc.abstractmethod
    def list_shards(self, split: str) -> Sequence[str]:
        """Returns string identifiers of input shards."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> Dataset:
        """Overrides base class to add shard identifier and remove use_cached.
        Args:
        split: string, the split to return.
        shuffle: bool, whether to shuffle the input source.
        seed: tf.int64 scalar tf.Tensor (or None) for shuffling input source.
        shard_info: optional specification for loading a shard of the split.
        """
        raise NotImplementedError

    def num_input_examples(self, split: str) -> Optional[int]:
        if self._num_input_examples is None:
            return None
        return self._num_input_examples[split]


    def _validate_args(fn, expected_pos_args):
        """Ensure function has exactly expected positional args."""
        argspec = inspect.getfullargspec(fn)
        expected_pos_args = tuple(expected_pos_args)
        actual_args = tuple(argspec.args)
        if actual_args[:len(expected_pos_args)] != expected_pos_args:
            raise ValueError("'%s' must have positional args %s, got: %s" % (fn.__name__, expected_pos_args, actual_args))
        actual_pos_args = tuple(argspec.args[:-len(argspec.defaults)] if argspec.defaults else argspec.args)
        if actual_pos_args != expected_pos_args[:len(actual_pos_args)]:
            raise ValueError("'%s' may only have positional args %s, got: %s" % (fn.__name__, expected_pos_args, actual_pos_args))


class TFDSDataSource(DataSource):
    """A `DataSource` that uses TensorFlow Datasets to provide the input data."""

    def __init__(
        self, 
        tfds_name: str, 
        tfds_data_dir: Optional[str] = None, 
        splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None
    ):
        """TfdsTask constructor.
        Args:
        tfds_name: string, the name and version number of a TFDS dataset,
            optionally with a config.
        tfds_data_dir: string, an optional path to a specific TFDS data directory
            to use.
        splits: an iterable of allowable string split names, a dict mapping
            allowable canonical splits (e.g., 'validation') to TFDS splits or slices
            (e.g., 'train[':1%']), or None. The default, None, uses all available
            splits from the TFDS dataset info.
        """
        if ":" not in tfds_name:
            raise ValueError(f"TFDS name must contain a version number, got: {tfds_name}")

        if splits and not isinstance(splits, dict):
            splits = {k: k for k in splits}

        self._tfds_dataset = LazyTFDSLoader(
            tfds_name, 
            data_dir = tfds_data_dir, 
            split_map = splits if isinstance(splits, dict) else None
        )

        # If splits are not provided, we pass an empty tuple and use the lazy
        # lookup in the `splits` property.
        super().__init__(splits=splits or ())

    @property
    def splits(self):
        """Overrides since we can't call `info.splits` until after init."""
        return self._splits or self._tfds_dataset.info.splits

    @property
    def tfds_dataset(self):
        return self._tfds_dataset

    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> Dataset:
        return self.tfds_dataset.load(split, shuffle_files=shuffle, seed=seed, shard_info=shard_info)

    def num_input_examples(self, split: str) -> int:
        """Overrides since we can't call `info.splits` until after init."""
        return self.tfds_dataset.size(split)

    def list_shards(self, split: str) -> Sequence[str]:
        return self.tfds_dataset.files(split)

class TFDSIODataSource(DataSource):
    def __init__(
        self, 
        config_or_file: Any, 
        preprocessors: Optional[Union[Callable, List[Callable]]] = None,
        splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None,
    ):
        self._tfds_dataset = LazyTFDSIOLoader(
            config_or_file = config_or_file, 
            preprocessors = preprocessors, 
            split_map = splits if isinstance(splits, dict) else None
        )
        super().__init__(splits = splits or ())
    
    @property
    def splits(self):
        """Overrides since we can't call `info.splits` until after init."""
        return self._tfds_dataset.info.splits

    @property
    def tfds_dataset(self):
        return self._tfds_dataset

    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> Dataset:
        return self.tfds_dataset.load(split, shuffle_files = shuffle, seed = seed, shard_info = shard_info)

    def num_input_examples(self, split: str) -> int:
        """Overrides since we can't call `info.splits` until after init."""
        return self.tfds_dataset.size(split)

    def list_shards(self, split: str) -> Sequence[str]:
        return self.tfds_dataset.files(split)


class HFDataSource(DataSource):
    def __init__(
        self, 
        config_or_file: Any, 
        dataset: Any,
        splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None,
    ):
        self._tfds_dataset = LazyHFTFDSLoader(
            config_or_file = config_or_file, 
            dataset = dataset, 
            split_map = splits if isinstance(splits, dict) else None
        )
        super().__init__(splits = splits or ())
    
    @property
    def splits(self):
        """Overrides since we can't call `info.splits` until after init."""
        return self._tfds_dataset.info.splits

    @property
    def tfds_dataset(self):
        return self._tfds_dataset

    def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> Dataset:
        return self.tfds_dataset.load(split, shuffle_files = shuffle, seed = seed, shard_info = shard_info)

    def num_input_examples(self, split: str) -> int:
        """Overrides since we can't call `info.splits` until after init."""
        return self.tfds_dataset.size(split)

    def list_shards(self, split: str) -> Sequence[str]:
        return self.tfds_dataset.files(split)

# class T5DataSource(DataSource):
#     def __init__(self, config_or_file: Any, preprocessor: Optional[Any] = None, splits: Optional[Union[Iterable[str], Mapping[str, str]]] = None):
#         self._tfds_dataset = LazyTFDSIOLoader(config_or_file=config_or_file, preprocessor=preprocessor, split_map=splits if isinstance(splits, dict) else None)
#         super().__init__(splits=splits or ())

#     @property
#     def splits(self):
#         """Overrides since we can't call `info.splits` until after init."""
#         return self._tfds_dataset.info.splits

#     @property
#     def tfds_dataset(self):
#         return self._tfds_dataset

#     def get_dataset(self, split: str, shuffle: bool = True, seed: Optional[int] = None, shard_info: Optional[ShardInfo] = None) -> tf.data.Dataset:
#         return self.tfds_dataset.load(split, shuffle_files=shuffle, seed=seed, shard_info=shard_info)

#     def num_input_examples(self, split: str) -> int:
#         """Overrides since we can't call `info.splits` until after init."""
#         return self.tfds_dataset.size(split)

#     def list_shards(self, split: str) -> Sequence[str]:
#         return self.tfds_dataset.files(split)



import abc
from typing import Any, Callable, Iterable, Mapping, MutableMapping, Optional, Sequence, Tuple, Type, Union
from tensorflow.python.data import Dataset
from .base import Feature, ShardInfo

class DatasetProviderBase(metaclass=abc.ABCMeta):
    """Abstract base for classes that provide a tf.data.Dataset."""

    @abc.abstractproperty
    def output_features(self) -> Mapping[str, Feature]:
        raise NotImplementedError

    @abc.abstractproperty
    def splits(self) -> Sequence[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_dataset(
        self, 
        sequence_length: int, 
        split: str, 
        use_cached: bool = False, 
        shuffle: bool = True,
        seed: Optional[int] = None, 
        shard_info: Optional[ShardInfo] = None, 
        num_epochs: int = 1
    ) -> Dataset:
        """Returns the requested tf.data.Dataset."""
        raise NotImplementedError

    @abc.abstractmethod
    def num_input_examples(self, split: str) -> int:
        raise NotImplementedError


class DatasetProviderRegistry(object):
    _REGISTRY: MutableMapping[str, DatasetProviderBase]
    _PROVIDER_TYPE: Type[DatasetProviderBase]

    @classmethod
    def add_provider(cls, name: str, provider):
        """Adds a data provider instance to the registry."""
        if name in cls._REGISTRY:
            raise ValueError(f"Attempting to register duplicate provider: {name}")
        if not isinstance(provider, cls._PROVIDER_TYPE):
            raise ValueError(f"Attempting to register a class not of an invalid type. Expecting instance of {cls._PROVIDER_TYPE}, got {type(provider).__name__}")
        cls._REGISTRY[name] = provider

    @classmethod
    def add(cls, name: str, provider_cls, *provider_args, **provider_kwargs):
        """Instantiates and adds provider to the registry."""
        if not issubclass(provider_cls, cls._PROVIDER_TYPE):
            raise ValueError(f"Attempting to register a class not of an invalid type. Expecting instance of {cls._PROVIDER_TYPE}, got {provider_cls}")

        provider = provider_cls(*provider_args, **provider_kwargs)
        cls.add_provider(name, provider)
        return provider

    @classmethod
    def remove(cls, name):
        """Remove provider from the registry, if it exists."""
        if name in cls._REGISTRY:
            del cls._REGISTRY[name]

    @classmethod
    def get(cls, name):
        """Returns provider from the registry."""
        if name not in cls._REGISTRY:
            raise ValueError(f"Provider name not registered: {name}")
        return cls._REGISTRY[name]

    @classmethod
    def names(cls):
        """Returns all provider names in registry."""
        return cls._REGISTRY.keys()

    @classmethod
    def reset(cls):
        """Removes all of the registered tasks."""
        cls._REGISTRY = {}

    @classmethod
    def get_dataset(cls, name, sequence_length, split, use_cached=False, shuffle=True, seed=None, shard_info=None, num_epochs=1):
        """Returns the requested tf.data.Dataset."""
        return cls.get(name).get_dataset(sequence_length=sequence_length, split=split, use_cached=use_cached, shuffle=shuffle, seed=seed, shard_info=shard_info, num_epochs=num_epochs)



# Wrapper for Funcs
from typing import Union, Dict, Optional, List, Callable, Any
from .base import BuilderConfig, TFDSDatasetBuilder, HFDatasetBuilder
from tensorflow_datasets.core.as_dataframe import as_dataframe
from tensorflow_datasets.core.dataset_utils import as_numpy
# from .loaders import LazyTFDSIOLoader
from tfdsio.utils import logger


def tfds_dataset(
    config_or_file: Union[BuilderConfig, Dict[str, Any], str], 
    dataset: Optional[Any] = None,
    preprocessors: Optional[List[Callable]] = None, 
    build: bool = True, 
    to_tfdataset: bool = True, 
    to_numpy: bool = False, 
    to_df: bool = False, 
    **kwargs
):
    """
    Wrapper Function that quickly creates a TFDSIO Dataset
    """
    if isinstance(config_or_file, str, dict):
        config = BuilderConfig()
        config.from_auto(config_or_file)
    builder = HFDatasetBuilder(dataset = dataset, config = config, **kwargs) if dataset else TFDSDatasetBuilder(config = config, **kwargs)
    if preprocessors:
        builder.set_preprocessors(preprocessors)
    if build:
        build_args = kwargs.get('build_args', {})
        builder.download_and_prepare(**build_args)
    if to_tfdataset or to_numpy:
        dataset_args = kwargs.get('dataset_args', {})
        ds = builder.as_dataset(**dataset_args)
        if to_numpy:
            return as_numpy(ds)
        if to_df:
            return as_dataframe(ds, builder.info)
        return ds
    return builder


def tfds_sample(dataset, num_samples: int = 5, return_samples: bool = True):
    samples = []
    for ex in as_numpy(dataset.take(num_samples)):
        logger.info(ex)
        samples.append(ex)
    if return_samples:
        return samples


def hf_tfds_dataset(
    config_or_file: Union[BuilderConfig, Dict[str, Any], str], 
    dataset: Any,
    build: bool = True, 
    to_tfdataset: bool = True, 
    to_numpy: bool = False, 
    to_df: bool = False, 
    **kwargs
):
    """
    Wrapper Function that quickly creates a TFDSIO Dataset
    """
    if isinstance(config_or_file, str, dict):
        config = BuilderConfig()
        config.from_auto(config_or_file)

    builder = HFDatasetBuilder(dataset = dataset, config = config, **kwargs)
    if build:
        build_args = kwargs.get('build_args', {})
        builder.download_and_prepare(**build_args)
    if to_tfdataset or to_numpy:
        dataset_args = kwargs.get('dataset_args', {})
        ds = builder.as_dataset(**dataset_args)
        if to_numpy:
            return as_numpy(ds)
        if to_df:
            return as_dataframe(ds, builder.info)
        return ds
    return builder




import seqio
import functools
import seqio.utils
import t5.data.preprocessors
import t5.evaluation.metrics
from typing import List, Dict, Any, Optional

# Global Configuration Options


DEFAULT_PREPROCESSORS = [
    seqio.preprocessors.tokenize,
    seqio.CacheDatasetPlaceholder(),
    seqio.preprocessors.append_eos_after_trim,
]

DEFAULT_METRICS = [t5.evaluation.metrics.bleu, t5.evaluation.metrics.rouge]

PREPROCESSOR_FUNCS = {
    'unsupervised': t5.data.preprocessors.unsupervised,
    'random_spans_noise_mask': t5.data.preprocessors.random_spans_noise_mask,
    'span_corruption': t5.data.preprocessors.span_corruption,
    'iid_denoising': t5.data.preprocessors.iid_denoising,
    'lm': t5.data.preprocessors.lm,
    'prefix_lm': t5.data.preprocessors.prefix_lm,
    'random_split_text': t5.data.preprocessors.random_split_text,
}

LM_PREPROCESSORS = {
    'unsupervised_lm': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": None, "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.unsupervised,
        seqio.preprocessors.append_eos_after_trim
    ],
    'span_corruption_lm': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": None, "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.span_corruption,
        seqio.preprocessors.append_eos_after_trim
    ],
    'iid_denoising_lm': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": None, "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.iid_denoising,
        seqio.preprocessors.append_eos_after_trim
    ],
    'prefix_lm': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": None, "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.prefix_lm,
        seqio.preprocessors.append_eos_after_trim
    ]
}


PRETRAIN_PREPROCESSORS = {
    'unsupervised': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": "inputs", "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.unsupervised,
        seqio.preprocessors.append_eos_after_trim
    ],
    'span_corruption': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": "inputs", "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.span_corruption,
        seqio.preprocessors.append_eos_after_trim
    ],
    'iid_denoising': [
        functools.partial(
            t5.data.preprocessors.rekey, 
            key_map={"inputs": "inputs", "targets": "targets"}
        ),
        seqio.preprocessors.tokenize,
        seqio.CacheDatasetPlaceholder(),
        t5.data.preprocessors.iid_denoising,
        seqio.preprocessors.append_eos_after_trim
    ]
}


class TFDSIOConfig:
    tfds_data_dir_override: str = seqio.utils._TFDS_DATA_DIR_OVERRIDE
    global_cache_directories: List[str] = seqio.utils._GLOBAL_CACHE_DIRECTORIES

    default_preprocessors: List[Any] = DEFAULT_PREPROCESSORS
    lm_preprocessors: Dict[str, List[Any]] = LM_PREPROCESSORS
    pretrain_preprocessors: Dict[str, List[Any]] = PRETRAIN_PREPROCESSORS


    # Override TFDS Configuration
    @classmethod
    def set_tfds_data_dir_override(cls, tfds_data_dir_override: str):
        seqio.utils.set_tfds_data_dir_override(tfds_data_dir_override)

    @classmethod
    def set_global_cache_dirs(cls, global_cache_directories: List[str]):
        seqio.utils.set_global_cache_dirs(global_cache_directories)

    @classmethod
    def add_global_cache_dirs(cls, global_cache_directories: List[str]):
        seqio.utils.add_global_cache_dirs(global_cache_directories)
    
    @classmethod
    def create_lm_preprocessors(
        cls, 
        rekey_map: Dict[str, str] = {'inputs': None, 'targets': 'targets'}, 
        exclude_funcs: List[str] = None,
        include_funcs: List[str] = None,
    ) -> List[Any]:  # sourcery skip: default-mutable-arg
        """
        Create a list of preprocessors for language modeling tasks.
        If no exclude_funcs or include_funcs are provided, the default tasks are used.        
        """
        tasks = []
        for task_name in {'unsupervised', 'span_corruption', 'iid_denoising', 'prefix_lm'}:
            if include_funcs is not None and task_name not in include_funcs:
                continue
            if exclude_funcs is not None and task_name in exclude_funcs:
                continue
            task = []
            if rekey_map is not None:
                task.append(functools.partial(t5.data.preprocessors.rekey, key_map = rekey_map))
            task.extend(
                (
                    seqio.preprocessors.tokenize, 
                    seqio.CacheDatasetPlaceholder(), 
                    cls.lm_preprocessors[task_name], 
                    seqio.preprocessors.append_eos_after_trim
                )
            )
            tasks.append(task)
        return tasks




